@@ -217,6 +217,8 @@ class RedizConventions(MicroConventions):
     def leaderboard_name(self, name=None, delay=None):
         if name is None and delay is None:
             return self.LEADERBOARD[:-2]+'.json'
+        elif name is None:
+            return self.LEADERBOARD+str(delay)+'.json'
         else:
             return self.LEADERBOARD+name if delay is None else self.LEADERBOARD+str(delay)+self.SEP+name
 
